package de.ullisroboterseite.ursai2notification;

// Autor: https://UllisRoboterSeite.de

// Doku:  https://UllisRoboterSeite.de/android-AI2-notifier.html
// Created: 2020-09-07
//
// Version 1.0 (2020-09-07)
// -------------------------
// - Basis-Version
//
// Version 2.0 (2021-03-08)
// -------------------------
// - Ergänzt und komplett überarbeitet
// - Name ist jetzt UrsAI2Notification
//
// Version 2.1 (2021-03-16)
// -------------------------
// - Einige Methoden als public deklariert,
//   damit sie von KeepAlive aufgerufen werden können
//
// Version 2.2 (2021-03-20)
// -------------------------
// - Die Intent-ID wurde nicht korrekt weiter geleitet.
//
// Version 2.3 (2021-04-05)
// -------------------------
// - Die Anzeige der Notification-Eigenschaft NumberID stimmte nicht mit der internen Vorbelegung überein.
//
// Version 2.4 (2021-04-16)
// -------------------------
// - Eigenschaft Colorized hatte im Designer den falschen Datentyp.
// - Bei Benutzung von KeepAlive wurden bei der Intent-Art Event das Ereignis UrsNotification.OnClick nicht ausgelöst.
//
// Version 2.5 (2021-04-17)
// -------------------------
// - Wegen falscher Flags beim PendingIntent wurde das Ereignis OnClick nur einmal ausglöst.
//
// Version 2.6 (2021-05-13)
// -------------------------
// - ScreenToOpen hat bei Kodular nicht funktionert. Context.getPackageName() ist nicht nutzbar.
//   stattdessen: form.getClass().getPackage().getName()
//
// Version 2.7 (2021-05-15)
// -------------------------
// - ScreenToOpen hat bei Kodular nicht funktionert.
// - GetActiveNotifications hinzugefügt
//
// Version 2.8 (2021-06-13)
// -------------------------
// - FlagNewTask wurde nicht ausgewertet, Default auf true gesetzt
//
// Version 2.9 (2022-10-13)
// -------------------------
// Angepasst für SDK31 (Android 12):
// Alle PendingIntent erhalten das Flag FLAG_IMMUTABLE.
// Globale Exception wird gefangen und ins Log ausgegeben.
//
// Version 2.10 (2023-09-23)
// -------------------------
// Angepasst für SDK33 (Android 13):
// Um Notifications auszugeben, ist die Permission POST_NOTIFICATIONS notwendig
// AreNotificationsEnabled hinzugefügt
// OpenNotificationSettings hinzugefügt
// AreNotificationsPaused hinzugefügt

// Nützliche Links
// --------------------
// https://developer.android.com/reference/android/app/NotificationChannel
// https://developer.android.com/reference/android/app/Notification
// https://developer.android.com/reference/android/app/NotificationManager
// https://developer.android.com/reference/android/app/Notification.Builder
// https://developer.android.com/guide/topics/ui/notifiers/notifications
// https://developer.android.com/training/notify-user/build-notification

// https://github.com/jarlisson2/NotificationStyleAIX/blob/master/com/jdl/NotificationStyle.java
// https://community.kodular.io/t/free-notification-style-extension-with-various-types-of-notification/66981
// https://onesignal.com/notification-preview
// https://itnext.io/android-notification-bigpicturestyle-1f293e6cabaf
// https://www.youtube.com/watch?v=IHNC7OtOEI4
// https://stackoverflow.com/questions/12665131/how-to-detect-if-a-notification-has-been-dismissed

// https://www.xspdf.com/resolution/52517993.html

// https://commonsware.com/blog/Android/2010/05/26/html-tags-supported-by-textview.html
// http://daniel-codes.blogspot.com/2011/04/html-in-textviews.html

import java.util.*;

import javax.management.Notification;

import java.io.File;
import java.nio.file.Path;

import com.google.appinventor.components.annotations.*;
import com.google.appinventor.components.annotations.androidmanifest.*;
import com.google.appinventor.components.common.*;
import com.google.appinventor.components.runtime.*;
import com.google.appinventor.components.runtime.util.*;

import android.app.*;
import android.content.*;
import android.graphics.drawable.Icon;
import android.graphics.Bitmap;
import android.media.AudioAttributes;
import android.media.AudioAttributes.Builder;
import android.media.RingtoneManager;
import android.net.Uri;
import android.os.Handler;
import android.os.Build;
import android.provider.Settings;
import android.service.notification.StatusBarNotification;

import android.util.Log;

import android.content.pm.*;

import java.lang.reflect.*;

@DesignerComponent(version = 1 //
        , versionName = UrsAI2NotificationChannel.VersionName //
        , dateBuilt = UrsAI2NotificationChannel.dateBuilt //
        , description = "Wrapper for Android Notification." //
        , category = com.google.appinventor.components.common.ComponentCategory.EXTENSION //
        , nonVisible = true //
        , androidMinSdk = 11 //
        , helpUrl = "https://UllisRoboterSeite.de/android-AI2-notifier.html" //
        , iconName = "aiwebres/iconchannel.png")
@SimpleObject(external = true)
@UsesPermissions(permissionNames = "android.permission.READ_EXTERNAL_STORAGE, android.permission.VIBRATE"
        + ", com.android.alarm.permission.SET_ALARM, android.permission.WAKE_LOCK, android.permission.EXPAND_STATUS_BAR"
        + ", android.permission.REORDER_TASKS, android.permission.SCHEDULE_EXACT_ALARM, android.permission.POST_NOTIFICATIONS")
@UsesBroadcastReceivers(receivers = {
        @ReceiverElement(name = "de.ullisroboterseite.ursai2notification.AlarmReceiver") })

public class UrsAI2NotificationChannel extends ChannelProperties implements OnInitializeListener {
    static final String LOG_TAG = "NOTIFY";
    static final String VersionName = "2.10.0";
    static final String dateBuilt = "2023-10-20";
    final UrsAI2NotificationChannel thisInstance = this;

    public UrsAI2NotificationChannel(ComponentContainer container) {
        super(container);
        thisComponent = this;
        form.registerForOnInitialize(this);
    }

    @Override
    public void onInitialize() {
        createChannel();
    }

    @SimpleProperty(description = "Returns the component's version name.")
    public String Version() {
        return VersionName;
    }

    @SimpleProperty(description = "Returns running Android SDK version.")
    public int VersionSDK() {
        return Build.VERSION.SDK_INT;
    }

    // ================================================================================================
    @SimpleFunction(description = "Creates a Notification")
    public void CreateSimpleNotification(int NumberID, String Icon, String Title, String Text) {
        PendingIntent pendingIntent;
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M)
            pendingIntent = PendingIntent.getBroadcast(thisContext, 0, new Intent("UrsDoNothing"),
                    PendingIntent.FLAG_UPDATE_CURRENT | PendingIntent.FLAG_IMMUTABLE);
        else
            pendingIntent = PendingIntent.getBroadcast(thisContext, 0, new Intent("UrsDoNothing"), 0);

        Notification notification = UrsNotification.buildSimpleNotification(form, channelID, NumberID, Icon, Title,
                Text, true, pendingIntent);

        manager.notify(NumberID, notification);
    }

    // ================================================================================================
    @SimpleFunction(description = "Creates a Notification")
    public void CreateActionNotification(int NumberID, String Icon, String Title, String Text,
            Component UrsIntentObject) {
        if (!(UrsIntentObject instanceof UrsIntent)) {
            form.ErrorOccurred(this, "CreateActionNotification", 17006, "Invalid value at UrsIntentObject");
            return;
        }

        UrsIntent uio = (UrsIntent) UrsIntentObject;

        PendingIntent pendingIntent = uio.buildIntent(NumberID);
        Notification notification = UrsNotification.buildSimpleNotification(form, channelID, NumberID, Icon, Title,
                Text, true, pendingIntent);

        manager.notify(NumberID, notification);
    }

    // ================================================================================================
    @SimpleFunction(description = "Creates a Notification")
    public void CreateAiNotification(int NumberID, String Icon, String Title, String Text, String ScreenToOpen,
            String StartValue, boolean AutoCancel) {

        Intent notificationIntent = null;
        PendingIntent pendingIntent = null;

        String className = "";
        if (ScreenToOpen.isEmpty())
            ScreenToOpen = "Screen1";

        if (ScreenToOpen.contains(".")) // Komplette Angabe der Activity
            className = ScreenToOpen;
        else { // Screen dieser AI2-App.
            className = form.getClass().getPackage().getName() + "." + ScreenToOpen;
        }

        // Prüfen, ob die Klasse als Intent taugt.
        try {
            notificationIntent = new Intent(thisContext, Class.forName(className));
        } catch (Exception e) {
            form.ErrorOccurred(this, "Start", 17004, "Screen not found: " + ScreenToOpen);
            return;
        }

        notificationIntent.addFlags(
                Intent.FLAG_ACTIVITY_NEW_TASK + Intent.FLAG_ACTIVITY_MULTIPLE_TASK + Intent.FLAG_ACTIVITY_NO_ANIMATION);

        if (StartValue.isEmpty())
            StartValue = "NOTIFICATION" + NumberID;
        notificationIntent.putExtra("APP_INVENTOR_START", StartValue);

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M)
            pendingIntent = PendingIntent.getActivity(thisContext, 0, notificationIntent,
                    PendingIntent.FLAG_UPDATE_CURRENT | PendingIntent.FLAG_IMMUTABLE);
        else
            pendingIntent = PendingIntent.getActivity(thisContext, 0, notificationIntent,
                    PendingIntent.FLAG_UPDATE_CURRENT);

        Notification notification = UrsNotification.buildSimpleNotification(form, channelID, NumberID, Icon, Title,
                Text, AutoCancel, pendingIntent);

        manager.notify(NumberID, notification);
    }

    // ================================================================================================
    @SimpleFunction(description = "Creates a Notification")
    public void CreateNotification(Component UrsNotificationObject, Component UrsIntentObject) {
        if (!(UrsNotificationObject instanceof UrsNotification)) {
            form.ErrorOccurred(this, "CreateNotification", 17005, "Invalid value at UrsNotificationObject");
            return;
        }

        if (!(UrsIntentObject instanceof UrsIntent)) {
            form.ErrorOccurred(this, "CreateNotification", 17006, "Invalid value at UrsIntentObject");
            return;
        }

        UrsNotification uno = (UrsNotification) UrsNotificationObject;
        UrsIntent uio = (UrsIntent) UrsIntentObject;

        PendingIntent pendingIntent = uio.buildIntent(uno.numberID);
        Notification not = uno.buildNotification(channelID, importance, pendingIntent, soundEnabled, -1, -1, false);

        manager.notify(uno.numberID, not);
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Creates a notification alarm")
    public void CreateAlarm(int RequestCode, int Delay, int Type, boolean WakeUp, Component UrsNotificationObject,
            Component UrsIntentObject) {
        if (!(UrsNotificationObject instanceof UrsNotification)) {
            form.ErrorOccurred(this, "CreateAlarm", 17005, "Invalid value at UrsNotificationObject");
            return;
        }

        if (!(UrsIntentObject instanceof UrsIntent)) {
            form.ErrorOccurred(this, "CreateAlarm", 17006, "Invalid value at UrsIntentObject");
            return;
        }

        if (Delay < 1) {
            form.ErrorOccurred(this, "CreateAlarm", 17007, "Invalid Delay value");
            return;
        }

        createAlarmAt(RequestCode, System.currentTimeMillis() + Delay * 1000, Type, WakeUp, UrsNotificationObject,
                UrsIntentObject);
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Creates a notification alarm")
    public long CreateAlarmEx(int RequestCode, int Delay, int Type, boolean WakeUp, Component UrsNotificationObject,
            Component UrsIntentObject) {
        if (!(UrsNotificationObject instanceof UrsNotification)) {
            form.ErrorOccurred(this, "CreateAlarmEx", 17005, "Invalid value at UrsNotificationObject");
            return 0;
        }

        if (!(UrsIntentObject instanceof UrsIntent)) {
            form.ErrorOccurred(this, "CreateAlarmEx", 17006, "Invalid value at UrsIntentObject");
            return 0;
        }

        if (Delay < 1) {
            form.ErrorOccurred(this, "CreateAlarmEx", 17007, "Invalid Delay value");
            return 0;
        }

        return createAlarmAt(RequestCode, System.currentTimeMillis() + Delay * 1000, Type, WakeUp,
                UrsNotificationObject, UrsIntentObject);
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Creates a notification alarm")
    public void CreateAlarmAt(int RequestCode, long Millis, int Type, boolean WakeUp, Component UrsNotificationObject,
            Component UrsIntentObject) {
        if (!(UrsNotificationObject instanceof UrsNotification)) {
            form.ErrorOccurred(this, "CreateAlarmAt", 17005, "Invalid value at UrsNotificationObject");
            return;
        }

        if (!(UrsIntentObject instanceof UrsIntent)) {
            form.ErrorOccurred(this, "CreateAlarmAt", 17006, "Invalid value at UrsIntentObject");
            return;
        }

        if (Millis < System.currentTimeMillis()) {
            form.ErrorOccurred(this, "CreateAlarmAt", 17008, "Invalid Millis value");
            return;
        }
        createAlarmAt(RequestCode, Millis, Type, WakeUp, UrsNotificationObject, UrsIntentObject);
    }
    // ====================================================================================================================

    long createAlarmAt(int RequestCode, long Millis, int Type, boolean WakeUp, Component UrsNotificationObject,
            Component UrsIntentObject) {

        AlarmManager alarmManager = (AlarmManager) thisContext.getSystemService(thisContext.ALARM_SERVICE);

        UrsNotification ursNotificationObject = (UrsNotification) UrsNotificationObject;
        UrsIntent ursIntentObject = (UrsIntent) UrsIntentObject;

        // Intent für den Alarmmanager
        Intent alarmIntent = new Intent(thisContext, AlarmReceiver.class);

        alarmIntent = ursNotificationObject.addPropertiesToIntent(channelID, importance, alarmIntent);
        alarmIntent = ursIntentObject.addPropertiesToIntent(alarmIntent, RequestCode);

        PendingIntent pendingAlarmIntent;

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M)
            pendingAlarmIntent = PendingIntent.getBroadcast(thisContext, RequestCode, alarmIntent,
                    PendingIntent.FLAG_UPDATE_CURRENT | PendingIntent.FLAG_IMMUTABLE);
        else
            pendingAlarmIntent = PendingIntent.getBroadcast(thisContext, RequestCode, alarmIntent,
                    PendingIntent.FLAG_UPDATE_CURRENT);

        int alarmType = AlarmManager.RTC;
        if (WakeUp)
            alarmType = AlarmManager.RTC_WAKEUP;
        if (Type <= 0)
            Type = 0;
        if (Type >= 3)
            Type = 3;

        if (Type == 2 && Build.VERSION.SDK_INT < 23)
            Type = 0;
        if (Type == 3 && Build.VERSION.SDK_INT < 23)
            Type = 1;
        if (Type == 1 && Build.VERSION.SDK_INT < 19)
            Type = 0;

        if (Type <= 0)
            alarmManager.set(AlarmManager.RTC_WAKEUP, Millis, pendingAlarmIntent);
        if (Type == 1)
            // https://developer.android.com/reference/android/app/AlarmManager#setExact(int,%20long,%20android.app.PendingIntent)
            // ab Level 19
            alarmManager.setExact(AlarmManager.RTC_WAKEUP, Millis, pendingAlarmIntent);
        if (Type == 2)
            // https://developer.android.com/reference/android/app/AlarmManager#setAndAllowWhileIdle(int,%20long,%20android.app.PendingIntent)
            // ab Level 23
            alarmManager.setAndAllowWhileIdle(AlarmManager.RTC_WAKEUP, Millis, pendingAlarmIntent);
        if (Type >= 3)
            // https://developer.android.com/reference/android/app/AlarmManager#setExactAndAllowWhileIdle(int,%20long,%20android.app.PendingIntent)
            // ab Level 23
            alarmManager.setExactAndAllowWhileIdle(AlarmManager.RTC_WAKEUP, Millis, pendingAlarmIntent);

        return Millis;
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Creates a notification alarm")
    public void CreateProgressNotification(int CurrentProgress, int MaxProgress, boolean Indeterminate,
            Component UrsNotificationObject, Component UrsIntentObject) {

        if (!(UrsNotificationObject instanceof UrsNotification)) {
            form.ErrorOccurred(this, "CreateProgressNotification", 17005, "Invalid value at UrsNotificationObject");
            return;
        }

        if (!(UrsIntentObject instanceof UrsIntent)) {
            form.ErrorOccurred(this, "CreateProgressNotification", 17006, "Invalid value at UrsIntentObject");
            return;
        }

        if (CurrentProgress < 0)
            CurrentProgress = 0;
        if (MaxProgress < CurrentProgress)
            MaxProgress = CurrentProgress;

        UrsNotification uno = (UrsNotification) UrsNotificationObject;
        UrsIntent uio = (UrsIntent) UrsIntentObject;

        PendingIntent pendingIntent = uio.buildIntent(uno.numberID);

        manager.notify(uno.numberID, uno.buildNotification(channelID, importance, pendingIntent, soundEnabled,
                CurrentProgress, MaxProgress, Indeterminate));
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Cancels All Notifications")
    public void CancelAll() {
        manager.cancelAll();
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Cancels a Notification based on numberID")
    public void Cancel(int NumberID) {
        manager.cancel(NumberID);
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Hides the given notification channel.")
    public void HideChannel(String ChannelID) {
        if (Build.VERSION.SDK_INT < 26) { // Vor Version 26 gab es keine Kanäle
            return;
        }
        manager.deleteNotificationChannel(ChannelID);
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Cancels the given alarm.")
    public void CancelAlarm(int RequestCode) {
        AlarmManager alarmManager = (AlarmManager) form.getSystemService(form.ALARM_SERVICE);

        Intent intent = new Intent(form, AlarmReceiver.class);

        PendingIntent pendingIntent;
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M)
            pendingIntent = PendingIntent.getBroadcast(form, RequestCode, intent,
                    PendingIntent.FLAG_UPDATE_CURRENT | PendingIntent.FLAG_IMMUTABLE);
        else
            pendingIntent = PendingIntent.getBroadcast(form, RequestCode, intent,
                    PendingIntent.FLAG_UPDATE_CURRENT);

        alarmManager.cancel(pendingIntent);
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Open Home Screen.")
    public void OpenHomeScreen() {
        Intent home = new Intent("android.intent.action.MAIN");
        home.addCategory(Intent.CATEGORY_HOME);
        home.setFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
        form.startActivity(home);
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Closes the notification panel.")
    public void CloseSystemDialogs() {
        Intent it = new Intent(Intent.ACTION_CLOSE_SYSTEM_DIALOGS);
        form.sendBroadcast(it);

    }

    // ====================================================================================================================
    @SimpleFunction(description = "Opens the notification panel.")
    public void OpenNotificationPanel() {
        try {
            Object sbservice = form.getSystemService("statusbar");
            Class<?> statusbarManager = Class.forName("android.app.StatusBarManager");
            Method showsb;
            if (Build.VERSION.SDK_INT >= 17) {
                showsb = statusbarManager.getMethod("expandNotificationsPanel");
            } else {
                showsb = statusbarManager.getMethod("expand");
            }
            showsb.invoke(sbservice);
        } catch (Exception e) {
            Log.d(LOG_TAG, "OpenNotificationPanel: " + e.toString());
        }
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Checks if another instance of this app is already running.")
    public boolean AppIsAlreadyRunning() {
        int count = 0;
        ActivityManager activityManager = (ActivityManager) form.getSystemService(Context.ACTIVITY_SERVICE);
        List<ActivityManager.RunningTaskInfo> tasks = activityManager.getRunningTasks(Integer.MAX_VALUE);

        for (ActivityManager.RunningTaskInfo task : tasks) {
            if (form.getPackageName().equalsIgnoreCase(task.baseActivity.getPackageName()))
                count++;
        }
        return count > 1;
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Moves an already running instance of the app to the front.")
    public void MoveOtherAppToFront() {
        ActivityManager activityManager = (ActivityManager) form.getSystemService(Context.ACTIVITY_SERVICE);
        List<ActivityManager.RunningTaskInfo> tasks = activityManager.getRunningTasks(Integer.MAX_VALUE);

        for (ActivityManager.RunningTaskInfo task : tasks) {
            if (form.getPackageName().equalsIgnoreCase(task.baseActivity.getPackageName())) {
                if (form.getTaskId() != task.id) {
                    activityManager.moveTaskToFront(task.id, 0);
                    return;
                }
            }
        }
    }

    // ====================================================================================================================
    @SimpleFunction(description = "Reteurns a list of all active notification IDs.")
    public YailList GetActiveNotifications() {
        StatusBarNotification[] nots = manager.getActiveNotifications();
        Integer[] ids = new Integer[nots.length];

        for (int i = 0; i < nots.length; i++) {
            ids[i] = nots[i].getId();
        }
        return YailList.makeList(ids);
    }

    // ====================================================================================================================
    // https://developer.android.com/reference/android/provider/Settings#ACTION_CHANNEL_NOTIFICATION_SETTINGS
    // an API Level 26
    @SimpleFunction(description = "Opens the notification channel settings.")
    public void OpenNotificationSettings() {
        Intent intent = new Intent();
        intent.setAction("android.settings.APP_NOTIFICATION_SETTINGS");
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O && channelID != null) {
            intent.setAction(Settings.ACTION_CHANNEL_NOTIFICATION_SETTINGS);
            intent.putExtra(Settings.EXTRA_CHANNEL_ID, channelID);
            intent.putExtra("android.provider.extra.APP_PACKAGE", form.getPackageName());
        }
        intent.putExtra("app_package", form.getPackageName());
        intent.putExtra("app_uid", form.getApplicationInfo().uid);
        form.startActivity(intent);
    }

    // https://developer.android.com/reference/android/app/NotificationManager#areNotificationsEnabled()
    // ab API Level 24
    @SimpleProperty(description = "Returns whether notifications are enabled for this app.")
    public boolean AreNotificationsEnabled() {
        NotificationManager notificationManager = (NotificationManager) form
                .getSystemService(Context.NOTIFICATION_SERVICE);
        if (Build.VERSION.SDK_INT >= 24)
            return notificationManager.areNotificationsEnabled();
        else
            return true;
    }

    // https://developer.android.com/reference/android/app/NotificationManager#areNotificationsPaused()
    // ab API Level 29
    @SimpleProperty(description = "Returns whether notifications are paused for this app.")
    public boolean AreNotificationsPaused() {
        NotificationManager notificationManager = (NotificationManager) form
                .getSystemService(Context.NOTIFICATION_SERVICE);
        if (Build.VERSION.SDK_INT >= 29)
            return notificationManager.areNotificationsPaused();
        else
            return false;
    }

}